# pylint: disable=protected-access

from functools import wraps, total_ordering

def datatype(name, static_fields):
  """ Defines a base class for a datatype that will represent a row from the database,
      in an abstracted form.
  """
  @total_ordering
  class DataType(object):
    __name__ = name

    def __init__(self, **kwargs):
      self._db_id = kwargs.pop('db_id', None)
      self._inputs = kwargs.pop('inputs', None)
      self._fields = kwargs

      for name in static_fields:
        assert name in self._fields, 'Missing field %s' % name

    def __eq__(self, other):
      return self._db_id == other._db_id

    def __lt__(self, other):
      return self._db_id < other._db_id

    def __getattr__(self, name):
      if name in static_fields:
        return self._fields[name]

      raise AttributeError('Unknown field `%s`' % name)

    def __repr__(self):
      return '<%s> #%s' % (name, self._db_id)

  return DataType


def requiresinput(input_name):
  """ Marks a property on the data type as requiring an input to be invoked. """
  def inner(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
      if self._inputs.get(input_name) is None:
        raise Exception('Cannot invoke function with missing input `%s`' % input_name)

      kwargs[input_name] = self._inputs[input_name]
      result = func(self, *args, **kwargs)
      return result

    return wrapper
  return inner
